import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import PyQt5 as qt

'''
Most important piece of code, tests if testing is testing

Args:
    string (str): A string to test the testing function.
Returns:
    str: A string indicating that the testing function is working.
    

'''

def testingtesting(string):
    answ =  f"I'm testing this very much: {string}"
    return answ

